"""Exception classes used in the project."""


__all__ = ('UnmatchedNumberError', 'WrongAnswerError',
           'UnsupportedSyntaxError', 'RepeatedAnswerError')


class UnmatchedNumberError(Exception):
    """This error is raised when met an unmatched input set."""

    def __init__(self, user_input):
        """Initialization."""
        self.__user_input = str(sorted(user_input)).replace(',', '')

    def __repr__(self):
        """Print out."""
        return '你使用的数字%s与题目所给的不符' % (self.__user_input)

    def __str__(self):
        """Print out."""
        return self.__repr__()


class WrongAnswerError(Exception):
    """This error is raised when a user's answer is wrong."""

    def __init__(self, answer):
        """Initialization."""
        self.__answer = answer

    def __repr__(self):
        """Print out."""
        return '你的结果[%s]是错误的' % (self.__answer)

    def __str__(self):
        """Print out."""
        return self.__repr__()


class UnsupportedSyntaxError(Exception):
    """This error is raised when the parser can't recognize the syntax."""

    def __init__(self, info):
        """Initialization."""
        self.__info = info

    def __repr__(self):
        """Print out."""
        return '你的公式无法识别[%s]' % (self.__info)

    def __str__(self):
        """Print out."""
        return self.__repr__()


class RepeatedAnswerError(Exception):
    """This error is raised when an answer is repeated."""

    def __init__(self, repeated):
        """Initialization."""
        self.__repeated = repeated

    def __repr__(self):
        """Print out."""
        return '你的结果与解[%s]重复了' % (self.__repeated)

    def __str__(self):
        """Print out."""
        return self.__repr__()



"""Main module of this project."""

from .database import DATABASE_42
from .expr_utils import expr_eval, judge_equivalent
from .exceptions import (UnmatchedNumberError, WrongAnswerError,
                         UnsupportedSyntaxError, RepeatedAnswerError)
import random
import ast


class GameApp(object):
    """
    The main game.

    Available methods (+ means playing, - means not playing):
    __init__(): initialization. (Entry point)
    is_playing(): show the status of current game. (+-)
    generate_problem(): generate a problem from database. (+-)
    get_current_problem(): print current problem. (+)
    get_current_solved(): print current solutions. (+)
    get_current_solution_number(): print current solution number. (+)
    get_total_solution_number(): print total solution number. (+)
    start(): start the game. (-)
    stop(): stop the game. (+)
    solve(): put forward a solution. (+)
    """

    def __init__(self):
        """Start the game session, serving as an initialization."""
        self.__valid = []  # this list stores readable answers
        self.__formula = []  # this list stores converted formulas
        self.__playing = False  # this stores playing status

    def is_playing(self):
        """Incicate the game is started or not."""
        return self.__playing

    def generate_problem(self, minimum_solutions: int,
                         maximum_solutions: int) -> tuple:
        """Generate a random problem from the database."""
        problem_list = [
            k for k in DATABASE_42.keys()
            if minimum_solutions <= DATABASE_42[k] <= maximum_solutions
        ]
        problem = random.choice(problem_list)
        return problem

    def get_current_problem(self) -> str:
        """Get current problem. Effective when playing."""
        message = '本次42点的题目为: %d %d %d %d %d' % (
            self.__problem[0], self.__problem[1], self.__problem[2],
            self.__problem[3], self.__problem[4])
        return message

    def get_current_solved(self) -> str:
        """Get current valid solutions. Effective when playing."""
        line = []
        if len(self.__valid) > 0:
            line.append('有效求解:')
            total = 1
            for valid_expr in self.__valid:
                line.append('[%d] %s' % (total, valid_expr))
                total += 1
        else:
            line.append('当前暂无有效求解')
        message = ''
        for each_line in line:
            message = message + each_line + '\n'
        return message.strip()

    def get_current_solution_number(self) -> int:
        """Get the number of current solutions. Effective when playing."""
        return len(self.__valid)

    def get_total_solution_number(self) -> int:
        """Get the number of total solutions. Effective when playing."""
        return DATABASE_42[self.__problem]

    def __validate(self, math_expr: str) -> str:
        """Validate distinguishing expressions. Private method."""
        for ind in range(0, len(self.__formula)):
            curr_expr = self.__formula[ind]
            if judge_equivalent(self.__problem, math_expr, curr_expr):
                raise RepeatedAnswerError(self.__valid[ind])

    def solve(self, math_expr: str) -> str:
        """Put forward a solution."""
        math_expr = math_expr.replace(' ', '').replace('（',
                                                       '(').replace('）', ')')

        if len(math_expr) >= 30:
            raise UnsupportedSyntaxError('待识别的公式过长')

        try:
            expr_ast = ast.parse(math_expr, mode='eval').body
        except:
            raise UnsupportedSyntaxError('公式无法被解析')

        math_expr_value, simplified_expr, user_input_numbers = expr_eval(
            expr_ast, '', [])

        if math_expr_value != 42:
            raise WrongAnswerError(math_expr_value)

        if tuple(sorted(user_input_numbers)) != self.__problem:
            raise UnmatchedNumberError(user_input_numbers)

        self.__validate(simplified_expr)
        self.__formula.append(simplified_expr)
        self.__valid.append(math_expr)

    def start(self, minimum_solutions: int = 1,
              maximum_solutions: int = 100) -> str:
        """Start the game. Effective when not playing."""
        self.__problem = self.generate_problem(minimum_solutions,
                                               maximum_solutions)
        self.__valid = []
        self.__formula = []
        self.__playing = True

    def stop(self) -> str:
        """Stop the game. Effective when playing."""
        self.__playing = False



"""Deal with expressions."""

from fractions import Fraction
from .exceptions import UnsupportedSyntaxError
import ast
import random


def expr_eval(node, simplified, orig_num):
    """Evaluate a expression."""
    if isinstance(node, ast.BinOp):
        lval, left, orig_num = expr_eval(node.left, simplified, orig_num)
        op = node.op
        rval, right, orig_num = expr_eval(node.right, simplified, orig_num)

        # add parenthesis
        if isinstance(op, (ast.Mult, ast.Div)) and isinstance(
                node.left, ast.BinOp) and isinstance(node.left.op,
                                                     (ast.Add, ast.Sub)):
            left = '(%s)' % (left)

        if isinstance(op, (ast.Mult, ast.Div)) and isinstance(
                node.right, ast.BinOp) and isinstance(node.right.op,
                                                      (ast.Add, ast.Sub)):
            right = '(%s)' % (right)

        if isinstance(op, ast.Sub) and isinstance(
                node.right, ast.BinOp) and isinstance(node.right.op,
                                                      (ast.Add, ast.Sub)):
            right = '(%s)' % (right)

        if isinstance(op, ast.Div) and isinstance(
                node.right, ast.BinOp) and isinstance(node.right.op,
                                                      (ast.Mult, ast.Div)):
            right = '(%s)' % (right)

        # calculation and concatenation
        if isinstance(op, (ast.Add, ast.Sub)):
            if lval == 0:
                left = '0'
            if rval == 0:
                right = '0'

        if isinstance(op, (ast.Mult, ast.Div)):
            if lval == 1:
                left = '1'
            if rval == 1:
                right = '1'

        result, new_simplified = None, None
        if isinstance(op, ast.Add):
            result, new_simplified = lval + rval, left + '+' + right
        elif isinstance(op, ast.Sub):
            result, new_simplified = lval - rval, left + '-' + right
        elif isinstance(op, ast.Mult):
            result, new_simplified = lval * rval, left + '*' + right
        elif isinstance(op, ast.Div) and rval != 0:
            result, new_simplified = lval / rval, left + '/' + right
        elif isinstance(op, ast.Div) and rval == 0:
            raise UnsupportedSyntaxError('被除数为0')
        else:
            raise UnsupportedSyntaxError('不支持的运算符')
        return result, new_simplified, orig_num

    elif isinstance(node, ast.Num):
        number = node.n
        if type(number) is not int:
            raise UnsupportedSyntaxError('不支持浮点数计算')
        orig_num.append(number)
        if number not in [0, 1]:
            return Fraction(number), chr(number + 97), orig_num
        else:
            return Fraction(number), str(number), orig_num
    else:
        raise UnsupportedSyntaxError('不支持的运算符')


def judge_equivalent(problem, expr_1, expr_2):
    """Judge equivalent of two expressions."""
    count = 10
    random_number = random.sample(range(50000, 100000), len(problem) * count)
    for current in range(0, count):
        temp_expr_1, temp_expr_2 = expr_1, expr_2
        for i in range(0, len(problem)):
            temp_expr_1 = temp_expr_1.replace(
                chr(problem[i] + 97),
                str(random_number[i + current * len(problem)]))
            temp_expr_2 = temp_expr_2.replace(
                chr(problem[i] + 97),
                str(random_number[i + current * len(problem)]))
        temp_ast_1 = ast.parse(temp_expr_1, mode='eval').body
        temp_ast_2 = ast.parse(temp_expr_2, mode='eval').body
        if expr_eval(temp_ast_1, '',
                     [])[0] == expr_eval(temp_ast_2, '', [])[0]:
            return True
    return False
